/*
Copyright (c) 2007,2007 Elizeu Nogueira da R. Jr.

All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, 
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution.
    * Neither the name of the Elizeu Nogueira da R. Jr. nor the names of its 
      contributors may be used to endorse or promote products derived from this 
      software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


This package is based on HypersonicSQL, originally developed by Thomas Mueller.
and HSQLDB original developed by the HSQL Development Group
*/

package br.com.dreamsource.mobile.jmesql;

import br.com.dreamsource.mobile.jmesql.exceptions.SQLException;

class Constraint {
    final static int FOREIGN_KEY = 0, MAIN = 1, UNIQUE = 2;
    private int      iType;
    private int      iLen;

    private Table    tMain;
    private int      iColMain[];
    private Index    iMain;
    private Object   oMain[];

    private Table    tRef;
    private int      iColRef[];
    private Index    iRef;
    private Object   oRef[];

    Constraint(int type, Table t, int col[]) {
	iType = type;
	tMain = t;
	iColMain = col;
	iLen = col.length;
    }

    Constraint(int type, Table main, Table ref, int cmain[],
	       int cref[]) throws SQLException {
	iType = type;
	tMain = main;
	tRef = ref;
	iColMain = cmain;
	iColRef = cref;
	iLen = cmain.length;

	if (Trace.ASSERT) {
	    Trace.assert1(cmain.length == cref.length);
	}

	oMain = tMain.getNewRow();
	oRef = tRef.getNewRow();
	iMain = tMain.getIndexForColumns(cmain);
	iRef = tRef.getIndexForColumns(cref);
    }

    int getType() {
	return iType;
    }

    Table getMain() {
	return tMain;
    }

    Table getRef() {
	return tRef;
    }

    int[] getMainColumns() {
	return iColMain;
    }

    int[] getRefColumns() {
	return iColRef;
    }

    void replaceTable(Table old, Table n) throws SQLException {
	if (old == tMain) {
	    tMain = n;
	} else if (old == tRef) {
	    tRef = n;
	} else {
	    Trace.assert1(false, "could not replace");
	}
    }

    void checkInsert(Object row[]) throws SQLException {
	if (iType == MAIN || iType == UNIQUE) {
	    return;
	}
	for (int i = 0; i < iLen; i++) {
	    Object o = row[iColRef[i]];
	    if (o == null) {
		return;
	    }
	    oMain[iColMain[i]] = o;
	}

	Trace.check(iMain.find(oMain) != null,
		    Trace.INTEGRITY_CONSTRAINT_VIOLATION);
    }

    void checkDelete(Object row[]) throws SQLException {
	if (iType == FOREIGN_KEY || iType == UNIQUE) {
	    return;
	}

	for (int i = 0; i < iLen; i++) {
	    Object o = row[iColMain[i]];

	    if (o == null) {
		return;
	    }
	    oRef[iColRef[i]] = o;
	}

	Trace.check(iRef.find(oRef) == null,
		    Trace.INTEGRITY_CONSTRAINT_VIOLATION);
    }

    void checkUpdate(int col[], Result deleted,
		     Result inserted) throws SQLException {
	if (iType == UNIQUE) {
	    return;
	}

	if (iType == MAIN) {
	    if (!isAffected(col, iColMain, iLen)) {
		return;
	    }

	    Record r = deleted.rRoot;

	    while (r != null) {

		if (iMain.find(r.data) == null) {
		    checkDelete(r.data);
		}

		r = r.next;
	    }
	} else if (iType == FOREIGN_KEY) {
	    if (!isAffected(col, iColMain, iLen)) {
		return;
	    }

	    Record r = inserted.rRoot;

	    while (r != null) {
		checkInsert(r.data);

		r = r.next;
	    }
	}
    }

    private boolean isAffected(int col[], int col2[], int len) {
	if (iType == UNIQUE) {
	    return false;
	}

	for (int i = 0; i < col.length; i++) {
	    int c = col[i];

	    for (int j = 0; j < len; j++) {
		if (c == col2[j]) {
		    return true;
		}
	    }
	}

	return false;
    }

}
